'''
Created on 28 Jul 2017

@author: julianporter
'''
import unittest
from .types import Generator
import traceback
VERBOSE=False


class TestFramework(unittest.TestCase):
    
    testName='test'
    
    def setUp(self):
        self.generator=Generator()
        self.nTests=1000
        self.errors=[]
        self.good=0
        self.count=0
        self.crashes=0
        
    def testAction(self):
        for _ in range(self.nTests): 
            try:
                self.good+=self.action()
                self.count+=1
            except:
                traceback.print_exc()
                self.crashes+=1
        if len(self.errors)==0:
            print('{} tests : count={}, good = {}, crashes = {}'.format(self.nTests,self.count,self.good,self.crashes))
        else:
            print('{} tests : count={}, good = {}, minimum log10 error = {}, crashes = {}'.format(self.nTests,self.count,self.good,min(self.errors),self.crashes))
        
        self.assertEqual(self.good,self.count)
     
    def action(self):
        return 1
    
    def push(self,e,threshold=10.0):
        self.errors.append(e)
        if e>threshold : return 1
        else: return 0
           
    def tearDown(self):
        pass


'''
Created on 28 Jul 2017

@author: julianporter
'''
import unittest
from .baseTest import TestFramework
from .types import equalMatrices, equalVectors, vectorDiff, matrixDiff
import numpy as np
from OSGridConverter.cartesian import Transform, Matrix




class TestVectorEquality(TestFramework):
    
    testName='vector creation'
    
    def action(self):
        np, v=self.generator.randomVectorPair()
        if equalVectors(np, v): return 1
        else: return 0

class TestVectorNegation(TestFramework):
    
    testName='vector negation'
        
    def action(self):

        np, v=self.generator.randomVectorPair()
        if equalVectors(-np, -v): return 1
        else: return 0
        
class TestDotProduct(TestFramework):
    
    testName='dot product'
        
    def action(self):
        np1, v1=self.generator.randomVectorPair()
        np2, v2=self.generator.randomVectorPair()
        if np.dot(np1,np2) == v1 | v2 : return 1
        else: return 0
        
class TestVectorSum(TestFramework):
    
    testName='vector sum'
        
    def action(self):
        np1, v1=self.generator.randomVectorPair()
        np2, v2=self.generator.randomVectorPair()
        if equalVectors(np1+np2,v1+v2) : return 1
        else: return 0
        
class TestMatrixSum(TestFramework):
    
    testName='matrix sum'
        
    def action(self):
        np1, v1=self.generator.randomMatrixPair()
        np2, v2=self.generator.randomMatrixPair()
        if equalMatrices(np1+np2,v1+v2) : return 1
        else: return 0
        
class TestMatrixVectorProduct(TestFramework):
    
    testName='matrix vector product'
        
    def action(self):
        np1, v1=self.generator.randomMatrixPair()
        np2, v2=self.generator.randomVectorPair()
        pr1 = np1@np2
        pr2 = v1@v2
        e=vectorDiff(pr1, pr2)
        return self.push(e)
    
class TestMatrixMatrixProduct(TestFramework):
    
    testName='matrix matrix product'
        
    def action(self):
        np1, v1=self.generator.randomMatrixPair()
        np2, v2=self.generator.randomMatrixPair()
        pr1 = np.transpose(np1@np2)
        pr2 = v1@v2
        e=matrixDiff(pr1, pr2)
        return self.push(e)
        
class TestHelmertInversion(unittest.TestCase):
    
    testName='Helmert matrix inversion'
 
    def setUp(self):
        super(TestHelmertInversion,self).setUp()
        self.i3=Matrix.identity(n=3)
        
    def action(self):
        t=self.generator.randomList()
        r=self.generator.randomList()
        s=self.generator.randomScalar()
        h=Transform(t,r,s).mx
        i=h.inverse()
        pr1 = h@i
        e= matrixDiff(pr1, self.i3)
        return self.push(e)
          

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 27 Jul 2017

@author: julianporter
'''
import traceback
from numbers import Number
import logging

class OSGridError(Exception):
    
    def __init__(self,message,inner=None):
        super(OSGridError,self).__init__()
        self.message='OSGridConverter error: {}'.format(message)
        self.inner=inner
        self.traceback=traceback.format_stack()
        
    def __str__(self):
        return self.message
    
def areNumbers(args):
    return all([isinstance(x,Number) and not isinstance(x,bool) for x in args])

class Log(object):
    
    def __init__(self,level=None):
        self.logger=logging.getLogger()
        if level:
            handler=logging.StreamHandler()
            self.logger.setLevel(level)
        else:
            handler=logging.NullHandler()
        self.logger.addHandler(handler)
        
    def debug(self,*args):
        self.logger.debug(args[0].format(*args[1:]))
        
    def info(self,*args):
        self.logger.info(args[0].format(*args[1:]))
        
    def warning(self,*args):
        self.logger.warning(args[0].format(*args[1:]))
        
    def error(self,*args):
        self.logger.error(args[0].format(*args[1:]))
        
    def exception(self,e):
        self.logger.error('{} {}',e,traceback.format_exc())

log=Log()
    
        


'''
Created on 26 Jul 2017

@author: julianporter
'''
from math import sin,cos,tan, pow, sqrt, radians
from .ellipsoids import Airy1830

class OSDefaultGrid (object):
    F0   = 0.9996012717
    phi0 = radians(49)
    l0   = radians(-2)
    N0   = -100000
    E0   =  400000
    ellipsoid = Airy1830
    
    def __init__(self,phi,E):
        self.phi=phi
        self.E=E
        
        _c=cos(phi)
        _s=sin(phi)
        _t=tan(phi)
        e2=self.ellipsoid.e1
        v   = 1.0-e2*pow(_s,2)
        _nu = self.aF0()/sqrt(v)
        rho = self.aF0()*(1-e2)*pow(v,-1.5)
        eta1= _nu/rho
        eta2= eta1 - 1.0
        
        def ncs(cpower=1,spower=0):
            if spower==0:
                return _nu*pow(_c,cpower)
            else: 
                return _nu*pow(_c,cpower)*pow(_s,spower)
            
        def t(power):
            return pow(_t,power)
        
        def nu(power):
            return pow(_nu,power)
    
        def rhoNu(power):
            return rho*pow(_nu,power)

        self.transforms=dict()
        M=self.meridional(self.phi)
        sc=1.0/_c
        self.transforms['I']     = M + self.N0
        self.transforms['II']    = ncs(1,1)/2.0
        self.transforms['III']   = ncs(3,1)*(5-t(2)+9*self.n(2))/24.0
        self.transforms['IIIA']  = ncs(5,1)*(61-58*t(2)+t(4))/720.0
        self.transforms['IV']    = ncs(1,0)
        self.transforms['V']     = ncs(3,0)*(eta1-t(2))/6.0
        self.transforms['VI']    = ncs(5,0)*(5 - 18*t(2) + t(4) + 14*eta2 - 58*t(2)*eta2)/120.0
        self.transforms['VII']   = t(1)/(2*rhoNu(1))
        self.transforms['VIII']  = t(1)/(24*rhoNu(3))*(5+3*t(2)+eta2-9*t(2)*eta2)
        self.transforms['IX']    = t(1)/(720*rhoNu(5))*(61+90*t(2)+45*t(4))
        self.transforms['X']     = sc/_nu
        self.transforms['XI']    = sc/(6*nu(3)*(eta1+2*t(2)))
        self.transforms['XII']   = sc/(120*nu(5)*(5+28*t(2)+24*t(4)))
        self.transforms['XIIA']  = sc/(5040*nu(7)*(61+662*t(2)+1320*t(4)+720*t(6)))
    
    def __getitem__(self,key):
        return self.transforms[key]
       
    def rhoNu(self,power):
        return self.rho*pow(self.nu,power)
    
    def Nu(self,power):
        return pow(self.nu,power)
    
    def s(self,power):
        return pow(self._s,power)
    
    def c(self,power):
        return pow(self._c,power)
    
    def t(self,power):
        return pow(self._t,power)
    
    @classmethod
    def n(cls,power=1):
        return pow(OSDefaultGrid.ellipsoid.n,power)
    
    def gridToLatLong(self):
        dE = self.E-OSDefaultGrid.E0
        phi = self.phi - self['VII']*pow(dE,2) + self['VIII']*pow(dE,4) - self['IX']*pow(dE,6)
        l = OSDefaultGrid.l0 + self['X']*dE - self['XI']*pow(dE,3) + self['XII']*pow(dE,5) - self['XIIA']*pow(dE,7)
        return (phi,l)
    
    def latLongToGrid(self):
        dl = self.E-OSDefaultGrid.l0
        N = self['I'] + self['II']*pow(dl,2) + self['III']*pow(dl,4) + self['IIIA']*pow(dl,6)
        E = OSDefaultGrid.E0 + self['IV']*dl + self['V']*pow(dl,3) + self['VI']*pow(dl,5)
        return (E,N)
    
    @classmethod
    def meridional(cls,phi):
        return cls.ellipsoid.meridional(phi,cls.phi0)*cls.bF0()
    
    @classmethod
    def aF0(cls):
        return cls.ellipsoid.a*cls.F0
    
    @classmethod
    def bF0(cls):
        return cls.ellipsoid.b*cls.F0



'''
Created on 19 Jul 2017

@author: julianporter
'''

from .base import OSGridError, areNumbers
from .mapping import Data, transformation
from .cartesian import Cartesian


class LatLong (object):
    def __init__(self,*args,tag='WGS84'):
        
        if len(args)==0: 
            raise OSGridError('Initialisation of LatLong requires at least one argument')
        self.tag=tag
        self.datum=Data[tag]
        if isinstance(args[0],Cartesian):
            self.latitude, self.longitude = args[0].LatLong(self.datum.ellipsoid)
        elif len(args)>=2 and areNumbers(args[:2]):
            self.latitude, self.longitude = args
        else:
            OSGridError('Arguments to LatLong constructor must be Cartesian or lat,long')
            
    def transform(self,newTag):
        if self.tag==newTag:
            return self
        else:
            oldC = Cartesian(self)      
            newC = transformation(self.tag,newTag,oldC)
            return LatLong(newC,tag=newTag)
    
    def __str__(self):
        return '{:>+.5f}:{:>+.5f}'.format(self.latitude,self.longitude)
    
    


'''
Created on 27 Jul 2017

@author: julianporter
'''
"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='OSGridConverter',
    version='0.1.1',
    description='Library to convert co-ordinates between the (UK) Ordnance Survey National Grid and latitude/longitude',
    long_description=long_description,
    url='https://github.com/....',

    author='Julian Porter',
    author_email='julian@jpembedded.solutions',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'Topic :: Education',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux'
    ],

    # What does your project relate to?
    keywords='gis mapping coordinates grid-reference mapping ordiance-survey lat-long',

    packages=find_packages(),
    install_requires=[],
    python_requires='>=3.4',
    
    extras_require={},
    package_data={},
    data_files=[],
    entry_points={},
    test_suite='test',
    tests_require=['OSGridConverter']
)



from .grid import OSGridReference
from .latlong import LatLong
from .mapping import Tags

def latlong2grid(latitude,longitude,tag='WGS84'):
    ll=LatLong(latitude,longitude,tag=tag)
    return OSGridReference(ll)
    
def grid2latlong(grid,tag='WGS84'):
    gr=OSGridReference(grid)
    return gr.toLatLong(tag=tag)

